/*#######################################################################################
Connect AVR to MMC/SD 

Copyright (C) 2004 Ulrich Radig

Bei Fragen und Verbesserungen wendet euch per EMail an

mail@ulrichradig.de

oder im Forum meiner Web Page : www.ulrichradig.de


Dieses Programm ist freie Software. Sie knnen es unter den Bedingungen der 
GNU General Public License, wie von der Free Software Foundation verffentlicht, 
weitergeben und/oder modifizieren, entweder gem Version 2 der Lizenz oder 
(nach Ihrer Option) jeder spteren Version. 

Die Verffentlichung dieses Programms erfolgt in der Hoffnung, 
da es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, 
sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT 
FR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License. 

Sie sollten eine Kopie der GNU General Public License zusammen mit diesem 
Programm erhalten haben. 
Falls nicht, schreiben Sie an die Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
#######################################################################################*/

#include "main.h"
#include "mmc_msp430.h"

#if USE_MMC

/**
 *
 */
void spi_init(unsigned char speed)
{
	U0CTL |= SWRST;		/* reset uart */

	ME1 &= ~(UTXE0+URXE0);	/* disable uart */
	ME1 |=  USPIE0;		/* Enable USART0 SPI mode */
	U0CTL = MM+SYNC+CHAR;
	U0TCTL = TXEPT+STC+SSEL0+SSEL1+CKPH;
	U0RCTL = 0;
	
	if (speed == MMC_LOWSPEED)
		U0BR0 = 0x80;		// 500khz
	else
		U0BR0 = 0x04;		// 
	U0BR1 = 0x00;		/* (U1BR1+U1BR0)=division factor for the baudrate */
	U0MCTL = 0x00;		/* no modulation */

	U0CTL &= ~SWRST;
}

/**
 *
 */
unsigned char mmc_init()
{
	unsigned char b, Timeout = 0;
	int a;

	if (DEBUG) printf("mmc_init()\n");
	
	//Konfiguration des Ports an der die MMC/SD-Karte angeschlossen wurde

	/* SIMO0 */
        P3OUT |=  BIT1;		/* P4.3 default value */
        P3DIR &= ~BIT1;		/* P4.3 as input */
        P3SEL |=  BIT1;		/* P4.3 SPI option select */

	/* SOMI0 */
        P3OUT |=  BIT2;		/* P4.4 default value */
        P3DIR |=  BIT2;		/* P4.4 as output */
        P3SEL |=  BIT2;		/* P4.4 SPI option select */

	/* UCLK */
        P3OUT |=  BIT3;		/* P4.5 default value */
        P3DIR |=  BIT3;		/* P4.5 as output */
        P3SEL |=  BIT3;		/* P4.5 SPI option select */

	P5OUT |=  BIT0;
        P5DIR |=  BIT0;		/* P5.0 output direction, used for Chip Select */
        P5SEL &= ~BIT0;		/* P5.0 function is output port */

	// sbi(MMC_Direction_REG,SPI_SS);	// niet nodig ???
	MMC_Disable();			//Setzt den Pin MMC_Chip_Select auf High Pegel

	// Wartet eine kurze Zeit
	for(a=0; a<200;a++){
		nop();
	};	

	// Aktiviren des SPI - Bus, Clock = Idel LOW
	// SPI Clock teilen durch 128
	spi_init(MMC_LOWSPEED);
	
	// init MMC/SD-card in SPI-mode
	for (b = 0;b<0x0f;b++) {	// send min 74+ Clocks to the MMC/SD-card
		Write_Byte_MMC(0xff);
	}
	
	//send command CMD0 to MMC/SD-card
	unsigned char CMD[] = {0x40,0x00,0x00,0x00,0x00,0x95};
	while(Write_Command_MMC(CMD) != 1) {
		if (Timeout++ > 200) {
			return(1);	// failed command 1 (Return code 1)
		}
	}
	Timeout = 0;
	CMD[0] = 0x41;	// command 1
	CMD[5] = 0xff;
	while(Write_Command_MMC(CMD) !=0) {
		if (Timeout++ > 100) {
			return(2);
		}
	}

	spi_init(MMC_HIGHSPEED);
	
	MMC_Disable();

	return(0);
}

/**
 *
 */
unsigned char Write_Command_MMC (unsigned char *CMD)
{
	unsigned char tmp = 0xff;
	unsigned char Timeout = 0;
	unsigned char a;

	/*if (DEBUG) {
		printf("Write_Command_MMC ");
		for (a=0;a<6;a++) printf("%02x ", CMD[a]);
		printf("\n");
	}*/
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv) 
	MMC_Disable();

	//sendet 8 Clock Impulse
	Write_Byte_MMC(0xFF);

	//set MMC_Chip_Select to low (MMC/SD-Karte Aktiv)
	MMC_Enable();

	//sendet 6 Byte Commando
	for (a = 0;a<0x06;a++) { //sendet 6 Byte Commando zur MMC/SD-Karte
		Write_Byte_MMC(*CMD++);
	}

	//Wartet auf ein gltige Antwort von der MMC/SD-Karte
	while (tmp == 0xff) {
		tmp = Read_Byte_MMC();
		if (Timeout++ > 100) {
			break; //Abbruch da die MMC/SD-Karte nicht Antwortet
		}
	}
	
	//if (DEBUG) printf("[%02x]", tmp);
	
	return(tmp);
}

/**
 *
 */
inline unsigned char Read_Byte_MMC (void)
{
	unsigned char dummy=0;
	
	//if (DEBUG) printf("Read_Byte_MMC()\n");

	dummy = U0RXBUF;	/* clear receive interrupt flag */
	while ((U0TCTL & TXEPT) == 0) {};
	U0TXBUF = 0xff;
	while ((IFG1 & URXIFG0) == 0) {};
	
	return U0RXBUF;
}

/**
 *
 */
inline void Write_Byte_MMC (unsigned char c)
{
	unsigned char dummy=0;
	
	//if (DEBUG) printf("Write_Byte_MMC()\n");

	dummy = U0RXBUF;	/* clear receive interrupt flag */
	while ((U0TCTL & TXEPT) == 0) {};
	U0TXBUF = c;
	while ((IFG1 & URXIFG0) == 0) {};
}

/**
 *
 */
unsigned char mmc_write_sector (unsigned long addr,unsigned char *Buffer)
{
	unsigned char tmp, a;
	//Commando 24 zum schreiben eines Blocks auf die MMC/SD - Karte
	unsigned char CMD[] = {0x58,0x00,0x00,0x00,0x00,0xFF}; 
	
	if (DEBUG) printf("mmc_write_sector()");

	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	  
	//addr = addr << 9; //addr = addr * 512
	
	CMD[1] = ((addr & 0xFF000000) >>15 );
	CMD[2] = ((addr & 0x00FF0000) >>7 );
	CMD[3] = ((addr & 0x0000FF00) <<1 );

	//Sendet Commando CMD24 an MMC/SD-Karte (Write 1 Block/512 Bytes)
	tmp = Write_Command_MMC (CMD);
	if (tmp != 0)
			{
			return(tmp);
			}
			
	//Wartet einen Moment und sendet einen Clock an die MMC/SD-Karte
	for (a=0;a<100;a++)
	{
	Read_Byte_MMC();
	}
	
	//Sendet Start Byte an MMC/SD-Karte
	Write_Byte_MMC(0xFE);	
	
	//Schreiben des Bolcks (512Bytes) auf MMC/SD-Karte
	for (a=0;a<512;a++)
		{
		Write_Byte_MMC(*Buffer++);
		}
	
	//CRC-Byte schreiben
	Write_Byte_MMC(0xFF); //Schreibt Dummy CRC
	Write_Byte_MMC(0xFF); //CRC Code wird nicht benutzt
	
	//Wartet auf MMC/SD-Karte Bussy
	while (Read_Byte_MMC() != 0xff){};
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	
	return(0);
}

/**
 *
 */
void MMC_Read_Block(unsigned char *CMD, unsigned char *Buffer, unsigned int n)
{	
	unsigned char q;
	unsigned int a;

	//Sendet Commando CMD an MMC/SD-Karte
	if (Write_Command_MMC(CMD) != 0) {
		return;
	}
			
	//Wartet auf Start Byte von der MMC/SD-Karte (FEh/Start Byte)
	while (Read_Byte_MMC() != 0xfe) {};
	
	//Lesen des Bolcks (normal 512Bytes) von MMC/SD-Karte
	for (a=0; a<n; a++) {
		*Buffer++ = Read_Byte_MMC();
	}
	
	//CRC-Byte auslesen
	Read_Byte_MMC();//CRC - Byte wird nicht ausgewertet
	Read_Byte_MMC();//CRC - Byte wird nicht ausgewertet
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	
	return;
}

/**
 *
 */
unsigned char mmc_read_sector(unsigned long addr, unsigned char *Buffer)
{	
	//Commando 16 zum lesen eines Blocks von der MMC/SD - Karte
	unsigned char CMD[] = {0x51,0x00,0x00,0x00,0x00,0xFF}; 
	
	if (DEBUG) printf("mmc_read_sector(%ld)\n", addr);

	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	  
	addr = addr << 9; //addr = addr * 512

	CMD[1] = ((addr & 0xFF000000) >>24 );
	CMD[2] = ((addr & 0x00FF0000) >>16 );
	CMD[3] = ((addr & 0x0000FF00) >>8 );

	MMC_Read_Block(CMD,Buffer,512);

	return(0);
}

/**
 *
 */
unsigned char Read_CID_MMC (unsigned char *Buffer)
{
	//Commando zum lesen des CID Registers
	unsigned char CMD[] = {0x4A,0x00,0x00,0x00,0x00,0xFF}; 
	
	if (DEBUG) printf("read_cid_mmc()\n");
	
	MMC_Read_Block(CMD,Buffer,16);

	return(0);
}

/**
 *
 */
unsigned char Read_CSD_MMC (unsigned char *Buffer)
{	
	// Command to read CSD register
	unsigned char CMD[] = {0x49,0x00,0x00,0x00,0x00,0xFF};
	
	if (DEBUG) printf("read_csd_mmc()\n");

	MMC_Read_Block(CMD,Buffer,16);

	return(0);
}

#endif //USE_MMC

